/*
===============================================================================

	FILE:	mesh.cpp
	
	CONTENTS:
	
		Face Fixer - SIGGRAPH 2000 demonstration software
	
	PROGRAMMERS:
	
		Martin Isenburg@cs.unc.edu
	
	COPYRIGHT:
	
		Copyright (C) 2000  Martin Isenburg (isenburg@cs.unc.edu)
		
		This software is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	
	CHANGE HISTORY:
	
		28 April 2000 -- Martin - finalized for SIGGRAPH submission.
	
===============================================================================
*/
#include "mesh.h"

#include "vector.h"
#include "vertex.h"
#include "normal.h"
#include "twinedge.h"
#include "material.h"
#include "group.h"
#include "vec3fv.h"

#include <string.h>

FFmesh::FFmesh()
{
  fs = new FFvector();
  es = new FFvector();
  vs = new FFvector();
  ns = new FFvector();
  ms = new FFvector();
  gs = new FFvector();
}

FFmesh::FFmesh(char* n)
{
  fs = new FFvector();
  es = new FFvector();
  vs = new FFvector();
  ns = new FFvector();
  ms = new FFvector();
  gs = new FFvector();
  strcpy(name, n);
}

FFmesh::~FFmesh()
{
  int i;

  FFtwinEdge* e;
  for (i = 0; i < es->size(); i++)
    {
      e = (FFtwinEdge*)es->elementAt(i);
      delete e;
    }
  FFvertex* v;
  for (i = 0; i < vs->size(); i++)
    {
      v = (FFvertex*)vs->elementAt(i);
      delete v;
    }
  FFnormal* n;
  for (i = 0; i < ns->size(); i++)
    {
      n = (FFnormal*)ns->elementAt(i);
      delete n;
    }
  FFmaterial* m;
  for (i = 0; i < ms->size(); i++)
    {
      m = (FFmaterial*)ms->elementAt(i);
      delete m;
    }
  FFgroup* g;
  for (i = 0; i < gs->size(); i++)
    {
      g = (FFgroup*)gs->elementAt(i);
      delete g;
    }
  delete fs;
  delete es;
  delete vs;
  delete ns;
  delete ms;
  delete gs;
}

void FFmesh::normalize()
{
  int i;

  float Min[3], Max[3], Trans[3], Dim[3], Scale;
  
  VecCopy3fv(Min,((FFvertex*)vs->elementAt(0))->v);
  VecCopy3fv(Max,((FFvertex*)vs->elementAt(0))->v);

  for (i = 1; i < vs->size(); i++)
    {
      VecUpdateMinMax3fv(Min,Max,((FFvertex*)vs->elementAt(i))->v);
    }

  VecSubtract3fv(Dim,Max,Min);
  if (Dim[0]>=Dim[1] && Dim[0]>=Dim[2]) Scale=2.0f/Dim[0];
  else if (Dim[1]>=Dim[0] && Dim[1]>=Dim[2]) Scale=2.0f/Dim[1];
  else Scale=2.0f/Dim[2];
  VecAdd3fv(Trans,Min,Max);
  VecSelfScalarMult3fv(Trans,0.5f);
  
  for (i = 0; i < vs->size(); i++)
    {
      VecSelfSubtract3fv( ((FFvertex*)vs->elementAt(i))->v, Trans );
      VecSelfScalarMult3fv( ((FFvertex*)vs->elementAt(i))->v, Scale );
    }
}

void FFmesh::computePerVertexNormals()
{
  int i;
  FFnormal* n;
  FFtwinEdge* e;
  FFtwinEdge* e_run;

  float tmpN[3];

  for (i = 0; i < ns->size(); i++)
  {
    n = (FFnormal*) ns->elementAt(i);
    delete n;
  }
  ns->removeAllElements();
  ns->ensureCapacity(vs->size());

  for (i = 0; i < es->size(); i++)
  {
    e = (FFtwinEdge*) es->elementAt(i);
    e->normal = 0;
  }

  for (i = 0; i < es->size(); i++)
  {
    e = (FFtwinEdge*) es->elementAt(i);
    if (e->group != -1)
      {
	if (e->normal == 0)
	  {
	    n = new FFnormal(0.0, 0.0, 0.0);
	    ns->addElement((void*) n);
	    while (e->normal == 0)
	      {
		e->normal = n;
		e = e->next;
	      }
	  }
	else
	  {
	    n = e->normal;
	  }
	e_run = e->next->inv->next->inv;
	while (e_run != e)
	  {
	    VecCcwNormal3fv(tmpN, e->vertex->v, e_run->vertex->v, e_run->inv->vertex->v);
	    VecSelfAdd3fv(n->n, tmpN);
	    e_run = e_run->next->inv;
	  }
      }
  }

  for (i = 0; i < ns->size(); i++)
  {
    n = (FFnormal*) ns->elementAt(i);    
    VecNormalize3fv(n->n);
  }
}

void FFmesh::computePerFaceNormals()
{
  int i;
  FFnormal* n;
  FFtwinEdge* e;
  FFtwinEdge* e_run;

  float tmpN[3];

  for (i = 0; i < ns->size(); i++)
  {
    n = (FFnormal*) ns->elementAt(i);
    delete n;
  }
  ns->removeAllElements();
  ns->ensureCapacity(fs->size());

  for (i = 0; i < es->size(); i++)
  {
    e = (FFtwinEdge*) es->elementAt(i);
    e->normal = 0;
  }

  for (i = 0; i < es->size(); i++)
  {
    e = (FFtwinEdge*) es->elementAt(i);
    if (e->group != -1)
      {
	if (e->normal == 0)
	  {
	    n = new FFnormal(0.0, 0.0, 0.0);
	    ns->addElement((void*) n);
	    while (e->normal == 0)
	      {
		e->normal = n;
		e = e->next->inv;
	      }
	  }
	else
	  {
	    n = e->normal;
	  }
	e_run = e->next->inv->next->inv;
	while (e_run != e)
	  {
	    VecCcwNormal3fv(tmpN, e->vertex->v, e_run->vertex->v, e_run->inv->vertex->v);
	    VecSelfAdd3fv(n->n, tmpN);
	    e_run = e_run->next->inv;
	  }
      }
  }

  for (i = 0; i < ns->size(); i++)
  {
    n = (FFnormal*) ns->elementAt(i);    
    VecNormalize3fv(n->n);
  }

  for (i = 0; i < es->size(); i++)
  {
    e = (FFtwinEdge*) es->elementAt(i);
    if (e->group == -1)
      {
	e->normal = e->next->normal;
      }
  }
}
